class FooService(object):

    def __init__(self, service_url, token_callback, ioloop):
        self.service_url = service_url
        self.token_callback = token_callback

    def assert_service_url_equals(self, expected_service_url):
        if not expected_service_url == self.service_url:
            raise AssertionError("Service URL mismatch.")

    async def assert_token_equals(self, expected_token):
        result = await self.token_callback()
        if expected_token != result:
            raise AssertionError("Token mismatch.")



import json

from tornado import gen
from tornado.httpclient import AsyncHTTPClient
from tornadorax import errors


class LoadBalancerService(object):

    def __init__(self, service_url, fetch_token, ioloop):
        self.service_url = service_url
        self.fetch_token = fetch_token
        self.ioloop = ioloop
        self.client = AsyncHTTPClient()

    @gen.coroutine
    def fetch_load_balancers(self):
        url = "{0}/loadbalancers".format(self.service_url)
        token = yield self.fetch_token()
        response = yield self.client.fetch(
            url, headers={"X-Auth-Token": token}, raise_error=False)
        errors.check_service_response(response, "loadbalancers")
        body = json.loads(response.body.decode("utf8"))
        load_balancers = []
        for lb_config in body["loadBalancers"]:
            load_balancers.append(
                LoadBalancer(lb_config["id"], lb_config["name"]))
        raise gen.Return(load_balancers)

    @gen.coroutine
    def fetch_load_balancer(self, lb_id):
        url = "{0}/loadbalancers/{1}".format(self.service_url, lb_id)
        token = yield self.fetch_token()
        response = yield self.client.fetch(
            url, headers={"X-Auth-Token": token}, raise_error=False)
        errors.check_service_response(response, "loadbalancers")
        body = json.loads(response.body.decode("utf8"))
        lb_config = body["loadBalancer"]
        lb = LoadBalancer(lb_config["id"], lb_config["name"])
        raise gen.Return(lb)

    @gen.coroutine
    def delete_load_balancer(self, lb_id):
        url = "{0}/loadbalancers/{1}".format(self.service_url, lb_id)
        token = yield self.fetch_token()
        response = yield self.client.fetch(
            url, method="DELETE", headers={"X-Auth-Token": token},
            raise_error=False)
        errors.check_service_response(response, "loadbalancers")

    @gen.coroutine
    def create_load_balancer(self, name, protocol, port, virtual_ips):
        url = "{0}/loadbalancers".format(self.service_url)
        token = yield self.fetch_token()
        body = json.dumps({
            "loadBalancer": {
                "name": name,
                "port": port,
                "protocol": protocol,
                "virtualIps": virtual_ips
            }
        })
        headers = {"X-Auth-Token": token, "Content-type": "application/json"}
        response = yield self.client.fetch(
            url, method="POST", headers=headers, body=body, raise_error=False)
        errors.check_service_response(response, "loadbalancers")
        lb_config = json.loads(response.body.decode("utf8"))["loadBalancer"]
        lb = LoadBalancer.from_config(lb_config)
        raise gen.Return(lb)


class LoadBalancer(object):

    @classmethod
    def from_config(cls, config):
        return cls(name=config["name"], lb_id=config["id"])

    def __init__(self, lb_id, name):
        self.id = lb_id
        self.name = name



import json
import logging
import hashlib
import hmac

try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from tornado.concurrent import Future
from tornado.httpclient import AsyncHTTPClient


CHUNK_SIZE = 64 * 1024
# sentinel value
READ_DONE = dict()


LOGGER = logging.getLogger("rax:storage")


class StorageService(object):

    def __init__(self, service_url, fetch_token, ioloop):
        self.service_url = service_url
        self.fetch_token = fetch_token
        self.ioloop = ioloop

    async def fetch_container(self, container_name):
        LOGGER.debug("Fetching container {0}".format(container_name))
        container_url = "{0}/{1}".format(self.service_url, container_name)
        container = StorageContainer(
            container_url, container_name, self.fetch_token,
            ioloop=self.ioloop)
        return container


class StorageContainer(object):

    def __init__(self, container_url, name, fetch_token, ioloop):
        self.name = name
        self.container_url = container_url
        self.fetch_token = fetch_token
        self.ioloop = ioloop

    async def fetch_object(self, object_name, tempurl_key=None):
        LOGGER.debug("Fetching object {0}".format(object_name))
        object_url = "{0}/{1}".format(self.container_url, object_name)
        storage_object = StorageObject(
            object_url, self.name, object_name, self.fetch_token, self.ioloop,
            tempurl_key=tempurl_key)
        return storage_object


class StorageObject(object):

    def __init__(
            self, url, container, object_name, fetch_token, ioloop,
            tempurl_key=None):
        self.object_url = url
        self.container = container
        self.name = object_name
        self.fetch_token = fetch_token
        self.ioloop = ioloop
        self.tempurl_key = tempurl_key
        self.client = AsyncHTTPClient()

    def generate_tempurl(self, method, expires):
        if not self.tempurl_key:
            raise MissingTempURLKey("'tempurl_key' parameter not provided.")
        expires = int(expires)
        path = urlparse.urlparse(self.object_url).path
        base = "{0}\n{1}\n{2}".format(method, expires, path).encode("utf8")
        key = self.tempurl_key.encode("utf8")
        signature = hmac.new(key, base, hashlib.sha1).hexdigest()
        params = urlencode({
            "temp_url_expires": str(expires),
            "temp_url_sig": signature
        })
        return self.object_url + "?" + params

    async def info(self):
        LOGGER.debug("Fetching object info: {0}".format(self.object_url))
        token = await self.fetch_token()
        headers = {"X-Auth-Token": token}
        response = await self.client.fetch(
            self.object_url, method="HEAD", headers=headers, raise_error=False)
        if response.code >= 400:
            return {
                "status": "error",
                "code": response.code,
                "body": response.body
            }
        metadata = {}
        values = {}
        for header, value in response.headers.items():
            if not header.startswith("X-"):
                continue
            if header.startswith("X-Object-Meta-"):
                name = header.split("X-Object-Meta-")[1].lower()
                metadata[name] = value
            else:
                values[header.lower()] = value

        values.update({
            "status": "success",
            "metadata": metadata,
            "type": response.headers["Content-type"],
            "length": int(response.headers["Content-length"]),
            "etag": response.headers["Etag"]
        })

        return values

    async def upload_stream(
            self, mimetype, writer=None, content_length=0, metadata=None):
        LOGGER.debug("Creating upload stream for {0}".format(self.object_url))
        metadata = metadata or {}
        extra_headers = dict([
            ("X-Object-Meta-{0}".format(key), value)
            for key, value in metadata.items()
        ])
        token = await self.fetch_token()
        writer = writer or BodyWriter
        writer_instance = writer(
            self.object_url, self.container, self.name, mimetype=mimetype,
            token=token, ioloop=self.ioloop, content_length=content_length,
            extra_headers=extra_headers)
        return writer_instance

    async def read(self, start=0, end=0):
        body = bytearray()
        reader = await self.read_stream(start=start, end=end)
        for read_future in reader:
            chunk = await read_future
            body.extend(chunk)
        return body

    async def read_stream(self, start=0, end=0):
        LOGGER.debug("Creating read stream {0}".format(self.object_url))
        token = await self.fetch_token()
        if end == 0:
            end = ""

        headers = {
            "X-Auth-Token": token,
            "Range": "bytes={0}-{1}".format(start, end)
        }

        chunks = []
        futures = []

        def body_callback(chunk):
            if futures:
                future = futures.pop(0)
                future.set_result(chunk)
            else:
                chunks.append(chunk)

        def response_callback(f):
            response = f.result()
            if response.code >= 400:
                LOGGER.debug("Error reading {0} ({1})".format(
                    self.object_url, response.code))
                exception = StreamError(
                    "Error retrieving object: {0}".format(response.code))
                if not futures:
                    raise exception
                future = futures.pop(0)
                future.set_exception(exception)
            else:
                if futures:
                    futures.pop(0).set_result("")
            LOGGER.debug("Finished reading {0}".format(self.object_url))
            chunks.append(READ_DONE)

        response_future = self.client.fetch(
            self.object_url, headers=headers,
            streaming_callback=body_callback, raise_error=False)

        response_future.add_done_callback(response_callback)

        def iterate():

            while True:
                future = Future()
                if not chunks:
                    if futures:
                        raise StreamError("Previous future was not consumed.")
                    futures.append(future)
                    yield future
                    continue

                chunk = chunks.pop(0)

                if chunk is READ_DONE:
                    break

                future.set_result(chunk)
                yield future

        return iterate()


class BodyWriter(object):

    def __init__(
            self, url, container_name, object_name, token, mimetype,
            ioloop, content_length, extra_headers=None):
        self.url = url
        self.content_length = content_length
        self.transferred_length = 0
        self.ioloop = ioloop
        self.md5sum = hashlib.md5()
        self.headers = {
            "X-Auth-Token": token,
            "Content-type": mimetype
        }
        self.headers.update(extra_headers or {})

        if content_length > 0:
            self.headers["Content-length"] = str(content_length)

        self.client = AsyncHTTPClient()
        self.initialized_future = Future()
        self.finish_future = Future()
        self.request_future = self.client.fetch(
            url, method="PUT", body_producer=self.body_producer,
            raise_error=False, headers=self.headers)

    async def body_producer(self, write_function):
        LOGGER.debug("Starting transfer to {0}".format(self.url))
        self.write_function = write_function
        self.initialized_future.set_result(None)
        await self.finish_future

    async def write(self, data):
        await self.initialized_future
        self.md5sum.update(data)
        self.transferred_length += len(data)
        await self.write_function(data)
        LOGGER.debug("Sent {0} to {1}".format(
            self.transferred_length, self.url))
        return len(data)

    async def finish(self):
        self.finish_future.set_result(None)
        LOGGER.debug("Closing file: {}".format(self.url))

        response = await self.request_future

        if response.code not in range(200, 300):
            LOGGER.debug("File transfer {0} failed: {1}".format(
                self.url, response.code))
            return {
                "status": "error",
                "code": response.code,
                "body": response.body
            }

        LOGGER.debug("Finished file: {}".format(self.url))

        return {
            "status": "success",
            "length": self.transferred_length,
            "md5sum": response.headers.get("ETag")
        }


# 1GB default segment size
DEFAULT_SEGMENT_SIZE = 1024 * 1024 * 1024


class SegmentWriter(object):

    @classmethod
    def with_defaults(cls, **cls_kwargs):
        def chunk_wrapper(*args, **kwargs):
            kwargs.update(cls_kwargs)
            return cls(*args, **kwargs)
        return chunk_wrapper

    def __init__(
            self, url, container, object_name, token, mimetype, ioloop,
            content_length, segment_size=DEFAULT_SEGMENT_SIZE, dynamic=False,
            extra_headers=None):
        self.url = url
        self.segment_size = segment_size
        self.container = container
        self.object_name = object_name
        self.token = token
        self.dynamic = dynamic
        self.mimetype = mimetype
        self.content_length = content_length
        self.ioloop = ioloop
        self.md5sum = hashlib.md5()
        self.segments = []
        self.segment_indexes = {}
        self.current_segment_number = 0
        self.current_segment_size = 0
        self.current_segment = None
        self.extra_headers = extra_headers or {}

    def create_segment(self, segment_name=None):
        if not segment_name:
            self.current_segment_number += 1
            segment_name = "%06d" % (self.current_segment_number)

        segment_name = "segments/{}".format(segment_name)
        segment_url = "{0}/{1}".format(self.url, segment_name)

        LOGGER.debug("Creating new segment {0}".format(segment_url))

        segment = BodyWriter(
            segment_url, self.container, self.object_name, self.token,
            "application/video-segment", self.ioloop, content_length=0)

        segment_path = "/{0}/{1}/{2}".format(
            self.container, self.object_name, segment_name)

        self.segment_indexes[id(segment)] = len(self.segments)

        self.segments.append({
            "path": segment_path,
            "etag": "",
            "size_bytes": 0
        })

        return segment

    async def write(self, data):
        if not self.current_segment:
            self.current_segment = self.create_segment()
            self.current_segment_size = 0

        chunk_size = \
            CHUNK_SIZE if CHUNK_SIZE < self.segment_size else self.segment_size

        if self.segment_size < self.current_segment_size + chunk_size:
            chunk_size = self.segment_size - self.current_segment_size

        if len(data) < chunk_size:
            chunk_size = len(data)

        await self.current_segment.write(data[:chunk_size])

        remaining_data = data[chunk_size:]
        self.current_segment_size += chunk_size

        if self.current_segment_size >= self.segment_size:
            await self.close_segment(self.current_segment)
            self.current_segment = None
            self.current_segment_size = 0

        if remaining_data:
            await self.write(remaining_data)

    async def close_segment(self, segment):
        result = await segment.finish()
        # TODO: verify and retry
        segment_index = self.segment_indexes[id(segment)]
        self.segments[segment_index]["etag"] = result["md5sum"]
        self.segments[segment_index]["size_bytes"] = result["length"]
        self.md5sum.update(result["md5sum"].encode("utf8"))
        return result

    async def finish(self):
        if self.current_segment:
            await self.close_segment(self.current_segment)

        client = AsyncHTTPClient()

        headers = {
            "X-Auth-Token": self.token,
            "Content-type": self.mimetype
        }

        headers.update(self.extra_headers)

        if self.dynamic:
            # dynamic manifests are much simpler, since they simply contain
            # a pattern to match, but they are eventually consistent with
            # updates and the segmentation is not explicit (meaning
            # that extra / duplicate segments are possible).
            headers["X-Object-Manifest"] = "{}/{}/segments".format(
                self.container, self.object_name)
            body = ""
            manifest_url = self.url
        else:
            # static manifests are more stable, since they include
            # explicit information about the chunks, but they do have size
            # limitations
            body = json.dumps(self.segments)
            manifest_url = self.url + "?multipart-manifest=put"

        response = await client.fetch(
            manifest_url, method="PUT", body=body, headers=headers)

        LOGGER.debug("Finished segmented delivery {0}".format(self.url))

        # TODO: verify and retry

        return {
            "etag": response.headers["Etag"],
            "status": "success",
            "md5sum": self.md5sum.hexdigest(),
            "length": sum([s["size_bytes"] for s in self.segments])
        }


class StreamError(Exception):
    pass


class MissingTempURLKey(Exception):
    pass



import datetime


FUTURE_EXPIRATION = datetime.datetime.utcnow() + \
    datetime.timedelta(days=60)


KEYSTONE_RESPONSE = {
    "access": {
        "token": {
            "expires": FUTURE_EXPIRATION.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "id": "TOKEN"
        },
        "serviceCatalog": [
            {
                "endpoints": [
                    {
                        "internalURL": "https://dfw.internal.com/v1",
                        "publicURL": "https://dfw.public.com/v1",
                        "region": "DFW"
                    },
                    {
                        "internalURL": "https://ord.internal.com/v1",
                        "publicURL": "https://ord.public.com/v1",
                        "region": "ORD"
                    }
                ],
                "type": "foo:service",
                "name": "Foo Service"
            },
            {
                "endpoints": [
                    {
                        "internalURL": "https://ord.queuest.com/v1",
                        "publicURL": "{{QUEUE_URL}}/v1",
                        "region": "ORD"
                    },
                    {
                        "internalURL": "https://dfw.queues.com/v1/111",
                        "publicURL": "https://dfw-int.queues.com/v1/111",
                        "region": "DFW"
                    }
                ],
                "name": "cloudQueues",
                "type": "rax:queues"
            }
        ]
    }
}



from tornado import gen
from tornado.testing import AsyncTestCase

from tests.helpers.foo_service import FooService
from tornadorax.services import service_registry

from importlib import reload


ENDPOINTS = [
    {
        "internalURL": "https://dfw1.foo.com/v1",
        "publicURL": "https://dfw2.foo.com/v1",
        "region": "DFW",
    },
    {
        "internalURL": "https://ord1.foo.com/v1",
        "publicURL": "https://ord2.foo.com/v1",
        "region": "ORD",
    }
]


@gen.coroutine
def fetch_token():
    raise gen.Return("TOKEN")


class TestServiceRegistry(AsyncTestCase):

    def setUp(self):
        super(TestServiceRegistry, self).setUp()
        service_registry.register_service("foo:service", FooService)
        self.addCleanup(lambda: reload(service_registry))

    def build(self, service_type="foo:service", **kwargs):
        service_catalog = [{
            "type": service_type,
            "name": service_type,
            "endpoints": ENDPOINTS
        }]
        kwargs.setdefault("service_catalog", service_catalog)
        kwargs.setdefault("fetch_token", fetch_token)
        kwargs.setdefault("ioloop", self.io_loop)
        return service_registry.build_service(service_type, **kwargs)

    def test_service_registry_includes_registered_services(self):
        service = self.build()
        # just select the first public URL without preferences
        service.assert_service_url_equals("https://dfw2.foo.com/v1")

    def test_service_registry_allows_internal_option(self):
        service = self.build(internal=True)
        service.assert_service_url_equals("https://dfw1.foo.com/v1")

    def test_service_registry_allows_region_option(self):
        service = self.build(internal=True, region="ORD")
        service.assert_service_url_equals("https://ord1.foo.com/v1")

    def test_service_registry_raises_error_with_invalid_region(self):
        with self.assertRaises(service_registry.InvalidRegion):
            self.build(internal=True, region="FOOBAR")

    def test_service_registry_raises_error_with_invalid_service(self):
        with self.assertRaises(service_registry.InvalidServiceType):
            self.build("other:service")

    def test_service_registry_raises_error_with_uncatalogued_service(self):
        with self.assertRaises(service_registry.InvalidServiceType):
            self.build(service_catalog=[])

    def test_service_registry_already_includes_core_services(self):
        self.build("rax:queues")
        self.build("object-store")
        self.build("rax:load-balancer")



import json
import hashlib
import random

try:
    from string import letters
except ImportError:
    from string import ascii_letters as letters

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from tornado.testing import AsyncTestCase, gen_test
from testnado.service_case_helpers import ServiceCaseHelpers

from tests.helpers.service_helpers import fetch_token
from tornadorax.services.storage_service import StorageService
from tornadorax.services.storage_service import SegmentWriter
from tornadorax.services.storage_service import MissingTempURLKey
from tornadorax.services.storage_service import StreamError


OBJECT_BODY = "".join([
    random.choice(letters) for i in range(2048)
]).encode("utf8")


class TestStorage(ServiceCaseHelpers, AsyncTestCase):

    def setUp(self):
        super(TestStorage, self).setUp()
        self.storage_service = self.add_service()
        self.storage_service.add_method(
            "PUT", "/v1/container/object", object_write_handle)

        self.storage_service.add_method(
            "GET", "/v1/container/object", object_read_handle)

        self.storage_service.add_method(
            "HEAD", "/v1/container/object", object_info_handle)

        self.storage_service.add_method(
            "PUT", "/v1/container/manifest", object_write_handle)

        self.storage_service.add_method(
            "PUT", r"/v1/container/manifest/segments/\d+", object_write_handle)

        self.client = StorageService(
            self.storage_service.url("/v1"), fetch_token=fetch_token,
            ioloop=self.io_loop)

    @gen_test
    async def test_generate_tempurl_requires_key(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object")
        with self.assertRaises(MissingTempURLKey):
            obj.generate_tempurl(method="GET", expires=1000)

    @gen_test
    async def test_generate_tempurl(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object", tempurl_key="foobar")
        url = obj.generate_tempurl(method="GET", expires=1000)
        url, params = url.split("?")
        self.assertEqual(url, obj.object_url)
        params = dict(urlparse.parse_qsl(params))
        # pregenerated based on parameters and 'foobar' key
        expected = "a42206ca0c6e654e46cd5e33b2b1f92aab81194d"
        self.assertEqual("1000", params["temp_url_expires"])
        self.assertEqual(expected, params["temp_url_sig"])

    @gen_test
    async def test_generate_tempurl_forces_integer_expiration(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object", tempurl_key="foobar")
        url = obj.generate_tempurl(method="GET", expires=1000.112)
        params = dict(urlparse.parse_qsl(url.split("?")[1]))
        # pregenerated based on parameters and 'foobar' key
        expected = "a42206ca0c6e654e46cd5e33b2b1f92aab81194d"
        self.assertEqual("1000", params["temp_url_expires"])
        self.assertEqual(expected, params["temp_url_sig"])

    @gen_test
    async def test_upload_stream_stores_contents(self):
        self.start_services()

        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object")
        writer = await obj.upload_stream(mimetype="text/html")
        await writer.write("CON".encode("ascii"))
        await writer.write("TENTS".encode("ascii"))
        result = await writer.finish()

        self.assertEqual("success", result["status"])
        self.assertEqual(8, result["length"])
        self.assertEqual(
            hashlib.md5(b"CONTENTS").hexdigest(), result["md5sum"])

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/object", headers={
                "X-Auth-Token": "TOKEN", "Content-type": "text/html"})

        self.assertEqual(b"CONTENTS", request.body)

    @gen_test
    async def test_upload_stream_allows_extra_metadata(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object")
        writer = await obj.upload_stream(
            mimetype="text/html", metadata={"foo": "bar", "cat": "mouse"})
        await writer.write(b"CONTENTS")
        result = await writer.finish()

        self.assertEqual("success", result["status"])
        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/object")

        self.assertEqual("bar", request.headers["X-Object-Meta-foo"])
        self.assertEqual("mouse", request.headers["X-Object-Meta-cat"])

    @gen_test
    async def test_upload_stream_raises_error(self):

        self.storage_service.add_method(
            "PUT", "/v1/container/object", object_write_error_handle)

        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object")
        writer = await obj.upload_stream(mimetype="text/html")
        await writer.write(b"CONTENTS")
        result = await writer.finish()

        self.assertEqual("error", result["status"])
        self.assertEqual(401, result["code"])
        self.assertEqual(b"ERROR", result["body"])

    @gen_test
    async def test_upload_stream_allows_content_length(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object")
        writer = await obj.upload_stream(
            mimetype="text/html", content_length=8)
        await writer.write(b"CONTENTS")
        await writer.finish()

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/object")
        self.assertEqual("8", request.headers["Content-length"])

    @gen_test
    async def test_upload_stream_allows_custom_metadata(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("manifest")
        segment_writer = SegmentWriter.with_defaults(segment_size=10)
        writer = await obj.upload_stream(
            mimetype="text/html", writer=segment_writer,
            metadata={"cat": "dog"})
        await writer.write(b"lincoln")
        result = await writer.finish()
        self.assertEqual("success", result["status"])

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest")
        self.assertEqual("dog", request.headers["X-Object-Meta-cat"])

    @gen_test
    async def test_upload_stream_allows_segmentation(self):
        # big, nasty segment test. should be broken up later, especially
        # with retry, etc.

        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("manifest")
        segment_writer = SegmentWriter.with_defaults(segment_size=4)
        writer = await obj.upload_stream(
            mimetype="text/html", writer=segment_writer)
        await writer.write(b"abe")
        await writer.write(b" lincoln")
        await writer.write(b" wins")
        result = await writer.finish()
        self.assertEqual("success", result["status"])

        expected = [
            ("abe ", "/container/manifest/segments/000001"),
            ("linc", "/container/manifest/segments/000002"),
            ("oln ", "/container/manifest/segments/000003"),
            ("wins", "/container/manifest/segments/000004")
        ]

        for content, segment_path in expected:
            request = self.storage_service.assert_requested(
                "PUT", "/v1{}".format(segment_path),
                headers={"X-Auth-Token": "TOKEN"})
            self.assertEqual(content, request.body.decode("utf8"))

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest",
            headers={"X-Auth-Token": "TOKEN"})
        self.assertEqual(b"put", request.arguments["multipart-manifest"][0])
        self.assertEqual("text/html", request.headers["Content-type"])

        body = json.loads(request.body.decode("utf8"))
        self.assertEqual(4, len(body))

        for i in range(len(body)):
            segment_info = body[i]
            expected_body, expected_path = expected[i]
            expected_etag = hashlib.md5(
                expected_body.encode("ascii")).hexdigest()
            self.assertEqual(expected_path, segment_info["path"])
            self.assertEqual(expected_etag, segment_info["etag"])
            self.assertEqual(4, segment_info["size_bytes"])

    @gen_test
    async def test_upload_stream_allows_custom_segments(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("manifest")
        # bad segment size so we can ensure it's not using it
        segment_writer = SegmentWriter.with_defaults(segment_size=1)
        writer = await obj.upload_stream(
            mimetype="text/html", writer=segment_writer)

        segment1 = writer.create_segment()
        await segment1.write(b"foo")
        await segment1.write(b"bar")
        await segment1.write(b"one")
        await writer.close_segment(segment1)
        segment2 = writer.create_segment()
        await segment2.write(b"foobar2")
        await writer.close_segment(segment2)
        result = await writer.finish()

        self.assertEqual("success", result["status"])

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest/segments/000001")
        self.assertEqual(b"foobarone", request.body)

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest/segments/000002")
        self.assertEqual(b"foobar2", request.body)

    @gen_test
    async def test_upload_segment_allows_dynamic_segments(self):
        self.start_services()

        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("manifest")

        writer = await obj.upload_stream(
            mimetype="text/html",
            writer=SegmentWriter.with_defaults(dynamic=True))

        segment1 = writer.create_segment("001")
        await segment1.write(b"foo")
        await writer.close_segment(segment1)

        segment2 = writer.create_segment("005")
        await segment2.write(b"bar")
        await writer.close_segment(segment2)

        result = await writer.finish()
        self.assertEqual("success", result["status"])

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest/segments/001")
        self.assertEqual(b"foo", request.body)

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest/segments/005")
        self.assertEqual(b"bar", request.body)

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest")
        self.assertEqual(
            "container/manifest/segments",
            request.headers["X-Object-Manifest"])
        self.assertEqual("text/html", request.headers["Content-type"])
        self.assertEqual(b"", request.body)

    # Need to add tests that verify etags, retry manifests, etc.

    @gen_test
    async def test_read_chunk(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object")
        chunk1 = await obj.read(0, 1023)
        chunk2 = await obj.read(1024)
        total = await obj.read()
        self.assertEqual(chunk1, OBJECT_BODY[:1024])
        self.assertEqual(chunk2, OBJECT_BODY[1024:])
        self.assertEqual(OBJECT_BODY, total)

    @gen_test
    async def test_read_stream_returns_body_in_chunks(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object")
        reader = await obj.read_stream()
        body = bytearray()
        for read_chunk in reader:
            chunk = await read_chunk
            body.extend(chunk)
        self.assertEqual(OBJECT_BODY, body)

    @gen_test
    async def test_read_stream_raises_with_uncomsumed_chunk(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object")
        reader = await obj.read_stream()
        next(reader)
        with self.assertRaises(StreamError):
            next(reader)

    @gen_test
    async def test_read_stream_raises_with_bad_response(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object2")
        reader = await obj.read_stream()
        with self.assertRaises(StreamError):
            for read_chunk in reader:
                await read_chunk

    @gen_test
    async def test_info_returns_metadata_about_object(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object")
        info = await obj.info()
        self.assertEqual("success", info["status"])
        self.assertEqual({"foo": "foo", "bar": "bar"}, info["metadata"])
        self.assertEqual("value", info["x-foobar"])
        self.assertEqual(1024, info["length"])
        self.assertEqual("text/plain", info["type"])
        self.assertEqual("md5sum", info["etag"])

    @gen_test
    async def test_info_returns_error_with_bad_response(self):
        self.start_services()
        container = await self.client.fetch_container("container")
        obj = await container.fetch_object("object2")
        info = await obj.info()
        self.assertEqual("error", info["status"])
        self.assertEqual(404, info["code"])


def object_write_handle(handler):
    handler.set_status(201)
    handler.set_header(
        "ETag", hashlib.md5(handler.request.body).hexdigest())
    handler.finish()


def object_write_error_handle(handler):
    handler.set_status(401)
    handler.write("ERROR")


def object_read_handle(handler):
    handler.set_status(200)
    range_string = handler.request.headers.get("Range", "bytes=0-")
    range_parts = range_string.split("=")[1].rsplit("-", 1)
    start, end = range_parts
    if not end:
        end = len(OBJECT_BODY) - 1
    start, end = (int(start), int(end) + 1)

    for i in range(0, end-start, 1024):
        offset = start + i
        offset_end = offset + 1024 if end > offset + 1024 else end
        handler.write(OBJECT_BODY[offset:offset_end])
        handler.flush()

    handler.finish()


def object_info_handle(handler):
    handler.set_status(200)
    handler.set_header("X-Object-Meta-Foo", "foo")
    handler.set_header("X-Object-Meta-Bar", "bar")
    handler.set_header("Etag", "md5sum")
    handler.set_header("Content-length", "1024")
    handler.set_header("Content-type", "text/plain")
    handler.set_header("X-Foobar", "value")
    handler.finish()



#!/usr/bin/env python

from setuptools import setup, find_packages


with open("VERSION") as version_fp:
    VERSION = version_fp.read().strip()

setup(
    name="tornadorax",
    version=VERSION,
    description="Tornado wrapper for OpenStack services",
    author='Josh Marshall',
    author_email='catchjosh@gmail.com',
    url="http://github.com/joshmarshall/tornadorax/",
    license="http://www.apache.org/licenses/LICENSE-2.0",
    packages=find_packages(exclude=["tests", "dist"]),
    install_requires=["tornado", "python-dateutil"]
)



import json

from tornado.testing import AsyncTestCase, gen_test
from testnado.service_case_helpers import ServiceCaseHelpers

from tests.samples import load_balancer_samples
from tests.helpers.service_helpers import fetch_token
from tornadorax import errors
from tornadorax.services.load_balancer_service import LoadBalancerService


class TestLoadBalancer(ServiceCaseHelpers, AsyncTestCase):

    def setUp(self):
        super(TestLoadBalancer, self).setUp()

        def handle_lb_list(handler):
            handler.write(load_balancer_samples.LIST_RESPONSE)

        def handle_lb_add(handler):
            handler.set_status(201)
            handler.write(load_balancer_samples.CREATE_RESPONSE)

        def handle_lb_get(handler):
            handler.set_status(200)
            handler.write(load_balancer_samples.GET_RESPONSE)

        def handle_lb_post(handler):
            handler.set_status(200)
            handler.write(load_balancer_samples.GET_RESPONSE)

        def handle_lb_delete(handler):
            handler.set_status(202)
            handler.finish()

        self.lb_service = self.add_service()
        self.lb_service.add_method(
            "GET", "/v1/account/loadbalancers", handle_lb_list)
        self.lb_service.add_method(
            "POST", "/v1/account/loadbalancers", handle_lb_add)
        self.lb_service.add_method(
            "GET", "/v1/account/loadbalancers/3", handle_lb_get)
        self.lb_service.add_method(
            "POST", "/v1/account/loadbalancers/3", handle_lb_post)
        self.lb_service.add_method(
            "DELETE", "/v1/account/loadbalancers/3", handle_lb_delete)

        self.client = LoadBalancerService(
            self.lb_service.url("v1/account"), fetch_token=fetch_token,
            ioloop=self.io_loop)

    @gen_test
    def test_load_balancer_fetch_returns_current_load_balancers(self):
        self.start_services()
        lbs = yield self.client.fetch_load_balancers()
        self.lb_service.assert_requested(
            "GET", "/v1/account/loadbalancers",
            headers={"X-Auth-Token": "TOKEN"})
        self.assertEqual(2, len(lbs))
        self.assertEqual("lb1", lbs[0].name)
        self.assertEqual(1, lbs[0].id)
        self.assertEqual("lb2", lbs[1].name)
        self.assertEqual(2, lbs[1].id)

    @gen_test
    def test_load_balancer_fetch_raises_service_error(self):
        self.start_services()
        self.lb_service.add_method(
            "GET", "/v1/account/loadbalancers",
            lambda x: x.set_status(500))
        with self.assertRaises(errors.ServiceError):
            yield self.client.fetch_load_balancers()

    @gen_test
    def test_load_balancer_fetch_returns_single_load_balancer(self):
        self.start_services()
        lb = yield self.client.fetch_load_balancer("3")
        self.lb_service.assert_requested(
            "GET", "/v1/account/loadbalancers/3",
            headers={"X-Auth-Token": "TOKEN"})

        self.assertEqual(lb.name, "a-new-loadbalancer")
        self.assertEqual(lb.id, 3)

    @gen_test
    def test_load_balancer_get_raises_service_error(self):
        self.start_services()
        self.lb_service.add_method(
            "GET", "/v1/account/loadbalancers/3",
            lambda x: x.set_status(500))
        with self.assertRaises(errors.ServiceError):
            yield self.client.fetch_load_balancer("3")

    @gen_test
    def test_load_balancer_delete_removes_single_load_balancer(self):
        self.start_services()
        yield self.client.delete_load_balancer("3")
        self.lb_service.assert_requested(
            "DELETE", "/v1/account/loadbalancers/3",
            headers={"X-Auth-Token": "TOKEN"})

    @gen_test
    def test_load_balancer_delete_raises_service_error(self):
        self.lb_service.add_method(
            "DELETE", "/v1/account/loadbalancers/3",
            lambda x: x.set_status(500))
        self.start_services()
        with self.assertRaises(errors.ServiceError) as exc_context:
            yield self.client.delete_load_balancer("3")
        self.assertEqual(500, exc_context.exception.code)

    @gen_test
    def test_load_balancer_create_builds_new_resource(self):
        self.start_services()
        lb = yield self.client.create_load_balancer(
            "a-new-loadbalancer", port=8888, protocol="HTTP",
            virtual_ips=[{"foo": "bar"}])
        request = self.lb_service.assert_requested(
            "POST", "/v1/account/loadbalancers",
            headers={
                "X-Auth-Token": "TOKEN",
                "Content-type": "application/json"
            })
        body = json.loads(request.body.decode("utf8"))["loadBalancer"]
        self.assertEqual("a-new-loadbalancer", body["name"])
        self.assertEqual(8888, body["port"])
        self.assertEqual("HTTP", body["protocol"])
        self.assertEqual([{"foo": "bar"}], body["virtualIps"])
        self.assertEqual("a-new-loadbalancer", lb.name)
        self.assertEqual(3, lb.id)

    @gen_test
    def test_load_balancer_create_raises_service_error(self):
        self.lb_service.add_method(
            "POST", "/v1/account/loadbalancers",
            lambda x: x.set_status(500))
        self.start_services()
        with self.assertRaises(errors.ServiceError) as exc_context:
            yield self.client.create_load_balancer(
                name="foo", protocol="foo", port=80, virtual_ips=[])
        self.assertEqual(500, exc_context.exception.code)
